#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <png.h>

#define _GINT_INTERNAL_INCLUDE_
#include "gint.h"
#include "gifont.h"

gi_font *gi_font_create( f32 ch_w, const char *png_file )
{
	gi_font *font = malloc( sizeof( gi_font ) );
	
	FILE *fp = fopen( png_file, "rb" );
	
	if ( fp == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] can't open font \"%s\" \n",
					__FILE__, __LINE__, png_file );
		return font;
	}
	
	ui8 sig[ 8 ];
	fread( sig, 8, 1, fp );
	
	if ( png_sig_cmp( sig, 0, 8 ) )
	{
		fprintf( stderr, "%s:%i [ERROR] \"%s\" isn't png file \n",
					__FILE__, __LINE__, png_file );
		return font;
	}
	
	png_structp png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
						NULL, NULL, NULL );
	if ( png_ptr == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] libpng failed \n",
					__FILE__, __LINE__ );
		return font;
	}
	
	png_infop info_ptr = png_create_info_struct( png_ptr );
	if ( info_ptr == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] libpng failed \n",
					__FILE__, __LINE__ );
		png_destroy_read_struct( &png_ptr, NULL, NULL );
		return font;
	}
	
	png_init_io( png_ptr, fp );
	png_set_sig_bytes( png_ptr, 8 );
	png_read_info( png_ptr, info_ptr );
	
	int i, w, h, bits, ltype;
	png_get_IHDR( png_ptr, info_ptr, ( void* ) &w, ( void* ) &h,
						&bits, &ltype, 0, 0, 0 );
	
	if ( bits > 8 )
		png_set_strip_16( png_ptr );
	if ( ltype == PNG_COLOR_TYPE_PALETTE )
		png_set_palette_to_rgb( png_ptr );
	if ( ( ltype == PNG_COLOR_TYPE_GRAY ) ||
		( ltype == PNG_COLOR_TYPE_GRAY_ALPHA ) )
	{
		if ( bits < 8 )
			png_set_gray_1_2_4_to_8( png_ptr );
		png_set_gray_to_rgb( png_ptr );
	}
	
	png_read_update_info( png_ptr, info_ptr );
	png_get_IHDR( png_ptr, info_ptr, ( void* ) &w, ( void* ) &h,
						&bits, &ltype, 0, 0, 0 );
	
	if ( ltype != PNG_COLOR_TYPE_RGB_ALPHA )
	{
		fprintf( stderr, "%s:%i [ERROR] \"%s\" haven't alpha channel\n",
					__FILE__, __LINE__, png_file );
		return font;
	}
	
	if ( w != h )
		fprintf( stderr, "%s:%i [WARNING] \"%s\" haven't quad sizes \n",
					__FILE__, __LINE__, png_file );
	
	void *data = malloc( w * h * 4 );
	uint *row_pointers = malloc( h * sizeof( uint* ) );
	
	for ( i = 0; i < h; i++ )
		row_pointers[ i ] = ( uint ) data + i * w * 4;
	
	png_read_image( png_ptr, ( void* ) row_pointers );
	
	png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
	free( row_pointers );
	fclose( fp );
	
	glGenTextures( 1, &font->tex_id );
	glBindTexture( GL_TEXTURE_2D, font->tex_id );
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, w, h, 0, GL_RGBA,
				GL_UNSIGNED_BYTE, data );
	
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	font->tex_size = w < h ? w : h;
	font->char_sw = ch_w;
	free( data );
	
	return font;
}

void gi_font_delete( gi_font *font )
{
	glDeleteTextures( 1, &font->tex_id );
	free( font );
}

gi_string *gi_string_create( gi_font *font, f32 size, const char *data )
{
	gi_string *string = malloc( sizeof( gi_string ) );
	
	uint i = 0;
	string->font = font;
	string->chars = strdup( data );
	string->length = strlen( string->chars );
	string->fsize = size;
	memset( string->fcl, 255, 4 );
	string->vdata = malloc( string->length * 4 * 16 );
	
	f32 fsize_x = string->fsize * string->font->char_sw;
	f32 fsize_tx = 0.03125f - 0.03125f * string->font->char_sw;
	
	for ( i = 0; i < string->length; i++ )
	{
		f32 fpsx = ( ( f32 ) i ) * fsize_x;
		f32 tlx = ( f32 ) ( ( ( unsigned char* ) string->chars ) [ i ] % 16 ) * 0.0625f;
		f32 tly = ( f32 ) ( ( ( unsigned char* ) string->chars ) [ i ] / 16 ) * 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 0 ) ) =
				fpsx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 4 ) ) =
				0.0f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 0 ) ) =
				fpsx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 4 ) ) =
				string->fsize;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 0 ) ) =
				fpsx + fsize_x;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 4 ) ) =
				string->fsize;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 0 ) ) =
				fpsx + fsize_x;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 4 ) ) =
				0.0f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 8 ) ) =
				tlx + fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 12 ) ) =
				tly;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 8 ) ) =
				tlx + fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 12 ) ) =
				tly + 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 8 ) ) =
				tlx + 0.0625f - fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 12 ) ) =
				tly + 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 8 ) ) =
				tlx + 0.0625f - fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 12 ) ) =
				tly;
	}
	
	return string;
}

void gi_string_update( gi_string *string, const char *data )
{
	uint old_length = string->length;
	string->length = strlen( data );

	if ( old_length < string->length )
	{
		string->vdata = realloc( string->vdata, string->length * 4 * 16 );
		string->chars = realloc( string->chars, string->length + 1 );
	}

	strcpy( string->chars, data );

	f32 fsize_x = string->fsize * string->font->char_sw;
	f32 fsize_tx = 0.03125f - 0.03125f * string->font->char_sw;

	uint i ;
	for ( i = 0; i < string->length; i++ )
	{
		f32 fpsx = ( ( f32 ) i ) * fsize_x;
		f32 tlx = ( f32 ) ( ( ( unsigned char* ) string->chars ) [ i ] % 16 ) * 0.0625f;
		f32 tly = ( f32 ) ( ( ( unsigned char* ) string->chars ) [ i ] / 16 ) * 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 0 ) ) =
				fpsx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 4 ) ) =
				0.0f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 0 ) ) =
				fpsx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 4 ) ) =
				string->fsize;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 0 ) ) =
				fpsx + fsize_x;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 4 ) ) =
				string->fsize;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 0 ) ) =
				fpsx + fsize_x;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 4 ) ) =
				0.0f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 8 ) ) =
				tlx + fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 0 ) * 16 + 12 ) ) =
				tly;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 8 ) ) =
				tlx + fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 1 ) * 16 + 12 ) ) =
				tly + 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 8 ) ) =
				tlx + 0.0625f - fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 2 ) * 16 + 12 ) ) =
				tly + 0.0625f;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 8 ) ) =
				tlx + 0.0625f - fsize_tx;
		* ( ( f32* ) ( ( uint ) string->vdata + ( i * 4 + 3 ) * 16 + 12 ) ) =
				tly;
	}
}

void gi_string_draw( gi_string *string )
{
	glBindTexture( GL_TEXTURE_2D, string->font->tex_id );

	glColor4ubv( string->fcl );

	glTexCoordPointer( 2, GL_FLOAT, 16,
				( void* ) ( ( uint ) string->vdata + 8 ) );
	glVertexPointer( 2, GL_FLOAT, 16, string->vdata );

	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	glEnableClientState( GL_VERTEX_ARRAY );

	glDrawArrays( GL_QUADS, 0, string->length * 4 );

	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState( GL_VERTEX_ARRAY );
}

void gi_string_draw_xy( gi_string *string, f32 x, f32 y )
{
	glPushMatrix();
	glTranslatef( x, y, 0.0f );

	gi_string_draw( string );

	glPopMatrix();
}

void gi_string_delete( gi_string *string )
{
	free( string->chars );
	free( string->vdata );
	free( string );
}

f32 gi_font_pxcalc( gi_font *font, uint px_size )
{
	uint width, height;
	gi_get( GI_WINDOW_WIDTH, &width );
	gi_get( GI_WINDOW_HEIGHT, &height );
	return ( f32 ) px_size / ( f32 ) height;
}

f32 gi_string_wcalc( gi_string *string )
{
	return ( ( f32 ) string->length ) * string->fsize *
				string->font->char_sw;
}

