#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define _GINT_INTERNAL_INCLUDE_
#include "gint.h"

#if defined(_GI_X11)

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/extensions/xf86vmode.h>
#include <GL/glx.h>

typedef struct {
	XVisualInfo		*visualinfo;
	Cursor			*cursor;
	Window  		window;
	GLXContext		glxc;
	int			vmid;	/* vidmode id to restore */
} gw_handle;

#endif /* _GI_X11 */

#if defined(_GI_WIN)

#include <windows.h>

typedef struct {
	HWND  			window;
	HDC			hDC;
	HGLRC			hRC;
	ui64			_tick;	/* crutch for */
	f64			_time;	/*	timer */
	int			_addx;	/* and for    */
	int			_addy;	/*  window ;) */
} gw_handle;

#endif /* _GI_WIN */

static gw_handle		*handle = NULL;
static gi_input			*input = NULL;

#if defined(_GI_X11)

#include <sys/time.h>

static Display			*display = NULL;

void gi_create()
{
	display = XOpenDisplay( NULL );
	if ( display == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] can't open default display \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	input = malloc( sizeof( gi_input ) );
	input->lkey = 0;
	memset( input->keys, GI_KEY_OFF, GI_KEYCOUNT * sizeof( ui8 ) );
	input->mouse_x = input->_mouse_x = 0;
	input->mouse_y = input->_mouse_y = 0;
}

void gi_delete()
{
	XCloseDisplay( display );
	free( input );
}

void gi_get( uint name, void *value )
{
	XWindowAttributes w_attr;
	switch ( name )
	{
		case GI_SCREEN_WIDTH:
			* ( int* ) value = ( int ) XDisplayWidth( display,
					XDefaultScreen( display ) );
			return ;

		case GI_SCREEN_HEIGHT:
			* ( int* ) value = ( int ) XDisplayHeight( display,
					XDefaultScreen( display ) );
			return ;

			//FIXME: 
		case GI_SCREEN_COLBPP:
			* ( int* ) value = 16;
			return ;

		case GI_WINDOW_WIDTH:
			XGetWindowAttributes( display, handle->window, &w_attr );
			* ( int* ) value = ( int ) w_attr.width;
			return ;

		case GI_WINDOW_HEIGHT:
			XGetWindowAttributes( display, handle->window, &w_attr );
			* ( int* ) value = ( int ) w_attr.height;
			return ;
	}
}

void gw_create( const char *title,
		int pos_x, int pos_y,
		int width, int height,
		gw_params *params )
{
	handle = ( gw_handle* ) malloc( sizeof( gw_handle ) );

	int glx_major, glx_minor;
	if ( glXQueryVersion( display, &glx_major, &glx_minor ) == FALSE )
	{
		fprintf( stderr, "%s:%i [PANIC] no glX extension \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	if ( params == NULL )
		fprintf( stderr, "%s:%i [WARNING] params undefined \n",
				__FILE__, __LINE__ );

	int screen = XDefaultScreen( display );
	Window root = XDefaultRootWindow( display );

	handle->vmid = -1;
	if ( ( params != NULL ) && ( params->flags & GW_FULLSCREEN ) )
	{
		int major, minor;
		if ( XF86VidModeQueryVersion( display, &major, &minor ) )
		{
			int i, vidmodes_n;
			XF86VidModeModeInfo **vidmodes;
			XF86VidModeGetAllModeLines( display,
					screen, &vidmodes_n, &vidmodes );

			int toid = -1, vmid = -1;
			int screen_width, screen_height;
			gi_get( GI_SCREEN_WIDTH, &screen_width );
			gi_get( GI_SCREEN_HEIGHT, &screen_height );

			for ( i = 0; i < vidmodes_n; i++ )
			{
				if ( ( vidmodes[ i ]->hdisplay == width ) &&
						( vidmodes[ i ]->vdisplay == height ) )
				{
					toid = i;
					break;
				}
			}

			if ( toid == -1 )
				fprintf( stderr, "%s:%i [ERROR] no suitable sizes \n",
						__FILE__, __LINE__ );

			for ( i = 0; i < vidmodes_n; i++ )
			{
				if ( ( vidmodes[ i ]->hdisplay == screen_width ) &&
						( vidmodes[ i ]->vdisplay == screen_height ) )
				{
					vmid = i;
					break;
				}
			}

			if ( vmid == -1 )
				fprintf( stderr, "%s:%i [ERROR] not found current vidmode \n",
						__FILE__, __LINE__ );

			if ( toid != vmid )
			{
				XF86VidModeSwitchToMode( display, screen, vidmodes[ toid ] );
				XF86VidModeSetViewPort( display, screen, 0, 0 );
				XSync( display, FALSE );
			}

			if ( toid != -1 )
				handle->vmid = vmid;
		}
		else
			fprintf( stderr, "%s:%i [ERROR] no XF86VidMode extension \n",
					__FILE__, __LINE__ );
	}

	int defcol = 0;
	gi_get( GI_SCREEN_COLBPP, &defcol );

	int attr[] = {	GLX_USE_GL,
		GLX_RGBA,
		GLX_BUFFER_SIZE, defcol,
		GLX_DEPTH_SIZE, 0,
		GLX_STENCIL_SIZE, 0,
		GLX_DOUBLEBUFFER,
		None };

	if ( params != NULL )
	{
		if ( params->cbits > 0 )
			attr[ 3 ] = params->cbits;

		if ( params->zbits > 0 )
			attr[ 5 ] = params->zbits;

		if ( params->sbits > 0 )
			attr[ 7 ] = params->sbits;
	}

	if ( ( handle->visualinfo = glXChooseVisual( display,
					XDefaultScreen( display ), attr ) ) == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] can't select visual \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	XSetWindowAttributes w_attr = {
		.colormap = XCreateColormap( display, root,
				handle->visualinfo->visual, AllocNone ),
		.border_pixel = 0,
		.background_pixel = 0,
		.override_redirect = ( handle->vmid != -1 ),
		.event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask |
			ButtonReleaseMask | EnterWindowMask |
			LeaveWindowMask | PointerMotionMask
	};

	XSync( display, FALSE );
	handle->window = XCreateWindow( display, root,
			pos_x, pos_y,
			width, height, 0,
			handle->visualinfo->depth,
			InputOutput,
			handle->visualinfo->visual,
			CWBorderPixel | CWColormap |
			CWOverrideRedirect | CWEventMask,
			&w_attr );

	XStoreName( display, handle->window, title );

	if ( ( params != NULL ) &&
			( params->flags & GW_HIDECURSOR ) )
	{
		XColor bk_color = {
			.pixel = 0,
			.red = 0, .green = 0, .blue = 0,
			.flags = 0,
			.pad = 0
		};

		handle->cursor = malloc( sizeof( Cursor ) );
		Pixmap pm_null = XCreateBitmapFromData( display,
				handle->window, "", 1, 1 );
		*handle->cursor = XCreatePixmapCursor( display, pm_null,
				pm_null, &bk_color, &bk_color, 0, 0 );
		XFreePixmap( display, pm_null );
		XDefineCursor( display, handle->window, *handle->cursor );
	}
	else
		handle->cursor = NULL;

	if ( params != NULL )
	{
		//FIXME: 
		if ( params->flags & GW_NOBORDER )
		{

		}
	}

	XSizeHints szhints = {
		.flags = PMinSize | PMaxSize,
		.min_width = width,
		.max_width = width,
		.min_height = height,
		.max_height = height
	};

	XSetWMNormalHints( display, handle->window, &szhints );
	XMapWindow( display, handle->window );

	if ( handle->vmid != -1 )
	{
		XMoveWindow( display, handle->window, 0, 0 );
		XWarpPointer( display, None,
				handle->window, 0, 0, 0, 0,
				width / 2, height / 2 );

		XGrabPointer( display, handle->window, FALSE,
				ButtonPressMask | ButtonReleaseMask |
				PointerMotionMask,
				GrabModeAsync, GrabModeAsync,
				handle->window,	None,
				CurrentTime );

		XGrabKeyboard( display, handle->window,	FALSE,
				GrabModeAsync, GrabModeAsync,
				CurrentTime );
	}

	XSync( display, FALSE );
	if ( ( handle->glxc = glXCreateContext( display, handle->visualinfo,
					None, TRUE ) ) == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] cannot create OpenGL context \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	XSync( display, FALSE );
	glXMakeCurrent( display, handle->window, handle->glxc );
	XSync( display, FALSE );

	if ( glXIsDirect( display, handle->glxc ) == FALSE )
		fprintf( stderr, "%s:%i [WARNING] direct rendering is disabled \n",
				__FILE__, __LINE__ );

	XSync( display, FALSE );
}

void gw_change_title( const char *title )
{
	XStoreName( display, handle->window, title );
}

void gw_delete()
{
	XUnmapWindow( display, handle->window );
	glXDestroyContext( display, handle->glxc );
	XDestroyWindow( display, handle->window );
	XFree( handle->visualinfo );

	if ( handle->cursor != NULL )
	{
		XFreeCursor( display, *handle->cursor );
		free( handle->cursor );
	}

	if ( handle->vmid != -1 )
	{
		XUngrabPointer( display, CurrentTime );
		XUngrabKeyboard( display, CurrentTime );
		int screen = XDefaultScreen( display );
		int vidmodes_n;
		XF86VidModeModeInfo **vidmodes;			
		XF86VidModeGetAllModeLines( display,
				screen, &vidmodes_n, &vidmodes );

		XF86VidModeSwitchToMode( display, screen, vidmodes[ handle->vmid ] );
		XF86VidModeSetViewPort( display, screen, 0, 0 );
	}

	XSync( display, FALSE );
	free( handle );
}

void gw_swap_buffers()
{
	glXSwapBuffers( display, handle->window );
	XFlush( display );
}

bool gi_is_ext( const char *ext )
{
	char *list = ( char* ) glGetString( GL_EXTENSIONS ), *is;
	if ( ( is = strstr( list, ext ) ) != NULL )
	{
		char *end = is + strlen( ext );
		if ( ( *end != '\040' ) ||
				( *end != '\024' ) ||
				( *end != '\000' ) )
			return TRUE;
		else
			return FALSE;
	}
	else
		return FALSE;
}

void *gi_ext_addr( const char *func_name )
{
	void *ptr = glXGetProcAddress( ( GLubyte* ) func_name );
	if ( ptr == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] adress of \"%s\" equale NULL",
				__FILE__, __LINE__, func_name );
		exit( EXIT_FAILURE );
	}

	return ptr;
}

gi_input *gi_input_update()
{
	int i, count = XEventsQueued( display, QueuedAlready )/* XPending( display )*/;

	for ( i = 0; i < count; i++ )
	{
		XEvent event;
		XNextEvent( display, &event );
		input->lkey = 0;

		switch ( event.type )
		{
			case KeyPress:
			case KeyRelease:
				//printf( "KEY: %c \n", (char)XLookupKeysym( &event.xkey, 0 ) );
				switch ( XLookupKeysym( &event.xkey, 0 ) )
				{
					case XK_A: case XK_a:
						input->keys[ GI_A ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_B: case XK_b:
						input->keys[ GI_B ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_C: case XK_c:
						input->keys[ GI_C ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_D: case XK_d:
						input->keys[ GI_D ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_E: case XK_e:
						input->keys[ GI_E ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F: case XK_f:
						input->keys[ GI_F ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_G: case XK_g:
						input->keys[ GI_G ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_H: case XK_h:
						input->keys[ GI_H ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_I: case XK_i:
						input->keys[ GI_I ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_J: case XK_j:
						input->keys[ GI_J ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_K: case XK_k:
						input->keys[ GI_K ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_L: case XK_l:
						input->keys[ GI_L ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_M: case XK_m:
						input->keys[ GI_M ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_N: case XK_n:
						input->keys[ GI_N ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_O: case XK_o:
						input->keys[ GI_O ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_P: case XK_p:
						input->keys[ GI_P ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Q: case XK_q:
						input->keys[ GI_Q ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_R: case XK_r:
						input->keys[ GI_R ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_S: case XK_s:
						input->keys[ GI_S ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_T: case XK_t:
						input->keys[ GI_T ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_U: case XK_u:
						input->keys[ GI_U ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_V: case XK_v:
						input->keys[ GI_V ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_W: case XK_w:
						input->keys[ GI_W ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_X: case XK_x:
						input->keys[ GI_X ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Y: case XK_y:
						input->keys[ GI_Y ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Z: case XK_z:
						input->keys[ GI_Z ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;

					case XK_0:
						input->keys[ GI_0 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_1:
						input->keys[ GI_1 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_2:
						input->keys[ GI_2 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_3:
						input->keys[ GI_3 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_4:
						input->keys[ GI_4 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_5:
						input->keys[ GI_5 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_6:
						input->keys[ GI_6 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_7:
						input->keys[ GI_7 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_8:
						input->keys[ GI_8 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_9:
						input->keys[ GI_9 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;

					case XK_F1:
						input->keys[ GI_F1 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F2:
						input->keys[ GI_F2 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F3:
						input->keys[ GI_F3 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F4:
						input->keys[ GI_F4 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F5:
						input->keys[ GI_F5 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F6:
						input->keys[ GI_F6 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F7:
						input->keys[ GI_F7 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F8:
						input->keys[ GI_F8 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F9:
						input->keys[ GI_F9 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F10:
						input->keys[ GI_F10 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F11:
						input->keys[ GI_F11 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_F12:
						input->keys[ GI_F12 ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;

					case XK_Escape:
						input->keys[ GI_ESC ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case 0:
						input->keys[ GI_TILD ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Tab:
						input->keys[ GI_TAB ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Shift_L: case XK_Shift_R:
						input->keys[ GI_SHIFT ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Control_L: case XK_Control_R:
						input->keys[ GI_CTRL ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Alt_L: case XK_Alt_R:
						input->keys[ GI_ALT ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_space:
						input->keys[ GI_SPACE ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Return:
						input->keys[ GI_ENTER ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_BackSpace:
						input->keys[ GI_BACKSPACE ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;

					case XK_Left:
						input->keys[ GI_LEFT ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Right:
						input->keys[ GI_RIGHT ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Up:
						input->keys[ GI_UP ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case XK_Down:
						input->keys[ GI_DOWN ] = ( event.type == KeyPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;

				}
				break;

			case ButtonPress:
			case ButtonRelease:
				switch ( event.xbutton.button )
				{
					case Button1:
						input->keys[ GI_BUTTON_LEFT ] =
							( event.type == ButtonPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case Button2:
						input->keys[ GI_BUTTON_MIDDLE ] =
							( event.type == ButtonPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case Button3:
						input->keys[ GI_BUTTON_RIGHT ] =
							( event.type == ButtonPress ) ? GI_KEY_DOWN : GI_KEY_UP;
						break;
					case Button4:
						input->keys[ GI_BUTTON_SCROLL_UP ] = GI_KEY_DOWN;
						break;
					case Button5:
						input->keys[ GI_BUTTON_SCROLL_DOWN ] = GI_KEY_DOWN;
						break;
				}
				break;

			case EnterNotify:
			case LeaveNotify:
				if ( ( input->_mouse_x == input->mouse_x ) &&
						( input->_mouse_y == input->mouse_y ) )
				{
					input->_mouse_x = input->mouse_x = ( int )
						event.xcrossing.x;
					input->_mouse_y = input->mouse_y = ( int )
						event.xcrossing.y;
				}
				else
					XWarpPointer( display, None,
							handle->window, 0, 0, 0, 0,
							input->_mouse_x = input->mouse_x,
							input->_mouse_y = input->mouse_y );
				break;

			case MotionNotify:
				if ( ( input->_mouse_x == input->mouse_x ) &&
						( input->_mouse_y == input->mouse_y ) )
				{
					input->_mouse_x = input->mouse_x = ( int )
						event.xmotion.x;
					input->_mouse_y = input->mouse_y = ( int )
						event.xmotion.y;
				}
				else
					XWarpPointer( display, None,
							handle->window, 0, 0, 0, 0,
							input->_mouse_x = input->mouse_x,
							input->_mouse_y = input->mouse_y );
				break;
		}
	}

	return input;
}

ui64 gi_timer_ui64()
{
	struct timeval time;
	gettimeofday( &time, NULL );
	return time.tv_usec + time.tv_sec * 1000000;
}

#endif /* _GI_X11 */

#if defined(_GI_WIN)

void gi_create()
{
	input = malloc( sizeof( gi_input ) );
	memset( input, 0, sizeof( gi_input ) );
}

void gi_delete()
{
	free( input );
}

void gi_get( uint name, void *value )
{
	RECT wrect;
	switch ( name )
	{
		case GI_SCREEN_WIDTH:
			* ( int* ) value = ( int ) GetSystemMetrics( SM_CXSCREEN );
			return ;

		case GI_SCREEN_HEIGHT:
			* ( int* ) value = ( int ) GetSystemMetrics( SM_CYSCREEN );
			return ;

		case GI_SCREEN_COLBPP:
			* ( int* ) value = 16;
			return ;

		case GI_WINDOW_WIDTH:
			GetWindowRect( handle->window, &wrect );
			* ( int* ) value = ( int ) ( wrect.right - wrect.left );
			return ;

		case GI_WINDOW_HEIGHT:
			GetWindowRect( handle->window, &wrect );
			* ( int* ) value = ( int ) ( wrect.bottom - wrect.top );
			return ;
	}
}

static LRESULT CALLBACK wnd_callback( HWND hWnd,
		UINT Msg, WPARAM wParam, LPARAM lParam )
{
	switch ( Msg )
	{
		case WM_CLOSE:
			break;

		case WM_KEYDOWN:
		case WM_KEYUP:
			switch ( wParam )
			{
				case VK_ESCAPE:
					input->keys[ GI_ESC ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_TAB:
					input->keys[ GI_TAB ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_SPACE:
					input->keys[ GI_SPACE ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_RETURN:
					input->keys[ GI_ENTER ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;

				case VK_LEFT:
					input->keys[ GI_LEFT ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_RIGHT:
					input->keys[ GI_RIGHT ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_UP:
					input->keys[ GI_UP ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case VK_DOWN:
					input->keys[ GI_DOWN ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;

				case 'A': case 'a':
					input->keys[ GI_A ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'B': case 'b':
					input->keys[ GI_B ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'C': case 'c':
					input->keys[ GI_C ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'D': case 'd':
					input->keys[ GI_D ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'E': case 'e':
					input->keys[ GI_E ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'F': case 'f':
					input->keys[ GI_F ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'G': case 'g':
					input->keys[ GI_G ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'H': case 'h':
					input->keys[ GI_H ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'I': case 'i':
					input->keys[ GI_I ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'J': case 'j':
					input->keys[ GI_J ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'K': case 'k':
					input->keys[ GI_K ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'L': case 'l':
					input->keys[ GI_L ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'M': case 'm':
					input->keys[ GI_M ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'N': case 'n':
					input->keys[ GI_N ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'O': case 'o':
					input->keys[ GI_P ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'P': case 'p':
					input->keys[ GI_P ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'Q': case 'q':
					input->keys[ GI_Q ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'R': case 'r':
					input->keys[ GI_R ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'S': case 's':
					input->keys[ GI_S ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'T': case 't':
					input->keys[ GI_T ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'U': case 'u':
					input->keys[ GI_U ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'W': case 'w':
					input->keys[ GI_W ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'X': case 'x':
					input->keys[ GI_X ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'Y': case 'y':
					input->keys[ GI_Y ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;
				case 'Z': case 'z':
					input->keys[ GI_Z ] = ( Msg == WM_KEYDOWN ) ? GI_KEY_DOWN : GI_KEY_UP;
					break;

			}
			break;

		case WM_MOUSEWHEEL:
			if ( ( ( i16 ) HIWORD( wParam ) ) > 0 )
				input->keys[ GI_BUTTON_SCROLL_UP ] = GI_KEY_DOWN;
			else
				input->keys[ GI_BUTTON_SCROLL_DOWN ] = GI_KEY_DOWN;
			break;

		case WM_LBUTTONDOWN:
			input->keys[ GI_BUTTON_LEFT ] = GI_KEY_DOWN;
			break;

		case WM_LBUTTONUP:
			input->keys[ GI_BUTTON_LEFT ] = GI_KEY_UP;
			break;

		case WM_MBUTTONDOWN:
			input->keys[ GI_BUTTON_MIDDLE ] = GI_KEY_DOWN;
			break;

		case WM_MBUTTONUP:
			input->keys[ GI_BUTTON_MIDDLE ] = GI_KEY_UP;
			break;

		case WM_RBUTTONDOWN:
			input->keys[ GI_BUTTON_RIGHT ] = GI_KEY_DOWN;
			break;

		case WM_RBUTTONUP:
			input->keys[ GI_BUTTON_RIGHT ] = GI_KEY_UP;
			break;

		case WM_MOUSEMOVE:
			// moved to gi_input_update()
			break;

		default:
			return DefWindowProc( hWnd, Msg, wParam, lParam );
	}

	return ( LRESULT ) NULL;
}

void gw_create( const char *title,
		int pos_x, int pos_y,
		int width, int height,
		gw_params *params )
{
	handle = ( gw_handle* ) malloc( sizeof( gw_handle ) );

	WNDCLASS w_class = {
		.style = CS_HREDRAW | CS_VREDRAW,
		.lpfnWndProc = &wnd_callback,
		.hInstance = ( HINSTANCE ) GetModuleHandle( NULL ),
		.hIcon = LoadIcon( 0, IDI_APPLICATION ),
		.hCursor = LoadCursor( 0, IDC_ARROW ),
		.hbrBackground = ( HBRUSH ) GetStockObject( BLACK_BRUSH ),
		.lpszMenuName = NULL,
		.lpszClassName = "gint"
	};

	if ( RegisterClass( &w_class ) == FALSE )
	{
		fprintf( stderr, "%s:%i [PANIC] RegisterClass() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	uint ws_flags = WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE;

	if ( params == NULL )
		fprintf( stderr, "%s:%i [WARNING] params undefined \n",
				__FILE__, __LINE__ );
	else
	{
		if ( params->flags & GW_NOBORDER )
			ws_flags |= WS_POPUP;
	}

	handle->window = CreateWindowEx( WS_EX_APPWINDOW, "gint", title,
			ws_flags, pos_x, pos_y, width, height,
			NULL, NULL, w_class.hInstance, NULL );

	if ( handle->window == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] CreateWindowEx() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	if ( ( handle->hDC = GetDC( handle->window ) ) == 0 )
	{
		fprintf( stderr, "%s:%i [PANIC] GetDC() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	uint pxformat ;
	PIXELFORMATDESCRIPTOR pfd;
	memset( &pfd, 0, sizeof( pfd ) );

	pfd.nSize = sizeof( pfd );
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER;
	pfd.dwLayerMask = PFD_MAIN_PLANE;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = ( params ) ? params->cbits : 16;
	pfd.cDepthBits = ( params ) ? params->zbits : 0 ;
	pfd.cAccumBits = 0;
	pfd.cStencilBits = params->sbits;

	if ( ( pxformat = ChoosePixelFormat( handle->hDC, &pfd ) ) == 0 )
	{
		fprintf( stderr, "%s:%i [PANIC] ChoosePixelFormat() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	if ( SetPixelFormat( handle->hDC, pxformat, &pfd ) == FALSE )
	{
		fprintf( stderr, "%s:%i [PANIC] SetPixelFormat() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	if ( ( handle->hRC = wglCreateContext( handle->hDC ) ) == 0 )
	{
		fprintf( stderr, "%s:%i [PANIC] wglCreateContext() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	if ( wglMakeCurrent( handle->hDC, handle->hRC ) == FALSE )
	{
		fprintf( stderr, "%s:%i [PANIC] wglMakeCurrent() failed \n",
				__FILE__, __LINE__ );
		exit( EXIT_FAILURE );
	}

	ShowWindow( handle->window, SW_NORMAL );

	if ( params != NULL )
	{
		if ( params->flags & GW_HIDECURSOR )
			ShowCursor( FALSE );

		if ( params->flags & GW_FULLSCREEN )
		{
			DEVMODE mode;
			EnumDisplaySettings( NULL, 0, &mode );

			mode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
			mode.dmSize = sizeof( mode );
			mode.dmBitsPerPel = ( params ) ? params->cbits : 16;
			mode.dmPelsWidth = width;
			mode.dmPelsHeight = height;

			if ( ChangeDisplaySettings( &mode, CDS_TEST ) == DISP_CHANGE_SUCCESSFUL )
			{
				ChangeDisplaySettings( &mode, CDS_FULLSCREEN );
				SetWindowPos( handle->window, HWND_TOP,	0, 0, width, height, SWP_SHOWWINDOW );
				ShowWindow( handle->window, SW_MAXIMIZE );
			}
		}
	}

	UpdateWindow( handle->window );

	//FIXME: need to clear
	handle->_time = 0.0;
	handle->_addx = /*GetSystemMetrics( SM_CXSIZEFRAME )*/ 0;
	handle->_addy = GetSystemMetrics( SM_CYCAPTION ) + GetSystemMetrics( SM_CYSIZEFRAME );
}

void gw_change_title( const char *title )
{
	//FIXME: 
}

void gw_delete()
{
	//FIXME: need to improve context deinit

	ChangeDisplaySettings( NULL, CDS_RESET );

	DestroyWindow( handle->window );
	UnregisterClass( "gint", GetModuleHandle( NULL ) );
}

void gw_swap_buffers()
{
	SwapBuffers( handle->hDC );
}

bool gi_is_ext( const char *ext )
{
	char *list = ( char* ) glGetString( GL_EXTENSIONS ), *is;
	if ( ( is = strstr( list, ext ) ) != NULL )
	{
		char *end = is + strlen( ext );
		if ( ( *end != '\040' ) ||
				( *end != '\024' ) ||
				( *end != '\000' ) )
			return TRUE;
		else
			return FALSE;
	}
	else
		return FALSE;
}

void *gi_ext_addr( const char *func_name )
{
	void *ptr = wglGetProcAddress( ( GLubyte* ) func_name );
	if ( ptr == NULL )
	{
		fprintf( stderr, "%s:%i [PANIC] adress of \"%s\" equale NULL",
				__FILE__, __LINE__, func_name );
		exit( EXIT_FAILURE );
	}

	return ptr;
}

gi_input *gi_input_update()
{
	MSG msg ;
	while ( PeekMessage( &msg, handle->window, 0, 0, PM_REMOVE ) )
	{
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}

	RECT rect ;
	POINT point;

	GetWindowRect( handle->window, &rect );
	GetCursorPos( &point );

	if ( ( input->_mouse_x == input->mouse_x ) &&
			( input->_mouse_y == input->mouse_y ) )
	{
		input->_mouse_x = input->mouse_x = ( point.x - rect.left + handle->_addx );
		input->_mouse_y = input->mouse_y = ( point.y - rect.top + handle->_addy );
	}
	else
	{
		int set_x = input->_mouse_x = input->mouse_x;
		int set_y = input->_mouse_y = input->mouse_y;
		set_x += rect.left - handle->_addx;
		set_y += rect.top - handle->_addy;
		SetCursorPos( set_x, set_y );
	}

	return input;
}

//FIXME: may use GetLocalTime ? 
ui64 gi_timer_ui64()
{
	ui64 tick, freq ;
	QueryPerformanceCounter( ( LARGE_INTEGER* ) &tick );
	QueryPerformanceFrequency( ( LARGE_INTEGER* ) &freq );
	handle->_time += ( ( f64 ) ( tick - handle->_tick ) ) / ( f64 ) freq;
	handle->_tick = tick;
	return ( ui64 ) ( handle->_time * 1000000.0 );
}

#endif /* _GI_WIN */

