#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <png.h>

#define _GINT_INTERNAL_INCLUDE_
#include "gint.h"
#include "gitools.h"

uint gi_img_load_png( const char *file, gi_raw_img *raw )
{
	FILE *fp = fopen( file, "rb" );
	
	if ( fp == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] can't open file \"%s\" \n",
						__FILE__, __LINE__, file );
		return 0;
	}
	
	ui8 sig[ 8 ];
	fread( sig, 8, 1, fp );
	
	if ( png_sig_cmp( sig, 0, 8 ) )
	{
		fprintf( stderr, "%s:%i [ERROR] \"%s\" is't png file \n",
						__FILE__, __LINE__, file );
		return 0;
	}
	
	png_structp png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING,
							NULL, NULL, NULL );
	if ( png_ptr == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] libpng failed \n",
						 __FILE__, __LINE__ );
		return 0;
	}
	
	png_infop info_ptr = png_create_info_struct( png_ptr );
	if ( info_ptr == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] libpng failed \n",
						 __FILE__, __LINE__ );
		png_destroy_read_struct( &png_ptr, NULL, NULL );
		return 0;
	}
	
	png_init_io( png_ptr, fp );
	png_set_sig_bytes( png_ptr, 8 );
	png_read_info( png_ptr, info_ptr );
	
	int w, h, bits, ltype;
	png_get_IHDR( png_ptr, info_ptr, ( void* ) &w, ( void* ) &h,
						&bits, &ltype, 0, 0, 0 );
	
	if ( bits > 8 )
		png_set_strip_16( png_ptr );
	
	if ( ltype == PNG_COLOR_TYPE_PALETTE )
		png_set_palette_to_rgb( png_ptr );
	
	if ( ( ltype == PNG_COLOR_TYPE_GRAY ) ||
		( ltype == PNG_COLOR_TYPE_GRAY_ALPHA ) )
	{
		if ( bits < 8 )
			png_set_gray_1_2_4_to_8( png_ptr );
		png_set_gray_to_rgb( png_ptr );
	}
	
	png_read_update_info( png_ptr, info_ptr );
	png_get_IHDR( png_ptr, info_ptr, ( void* ) &w, ( void* ) &h,
						&bits, &ltype, 0, 0, 0 );
	
	if ( ( ( ltype != PNG_COLOR_TYPE_RGB ) &&
		( ltype != PNG_COLOR_TYPE_RGB_ALPHA ) ) ||
		( bits > 8 ) )
	{
		fprintf( stderr, "%s:%i [ERROR] fixme: invalid png reading \n",
						__FILE__, __LINE__ );
	}
	
	bool alpha = ( ltype == PNG_COLOR_TYPE_RGB_ALPHA );
	int bytes = alpha ? 4 : 3 ;
	
	void *data = malloc( w * h * bytes );
	uint *row_pointers = malloc( h * sizeof( uint* ) );
	
	int i ;
	for ( i = 0; i < h; i++ )
		row_pointers[ i ] = ( uint ) data + i * w * bytes;
	
	png_read_image( png_ptr, ( void* ) row_pointers );
	
	png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
	free( row_pointers );
	fclose( fp );

	if ( raw != NULL )
	{
		raw->width = w;
		raw->height = h;
		raw->raw = ( ui8* ) data;
		raw->alpha = alpha;

		return ( 0 );
	}
	else
	{
		uint id ;
		glGenTextures( 1, &id );
		glBindTexture( GL_TEXTURE_2D, id );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

		if ( alpha )
			glTexImage2D( GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA, w, h,
					0, GL_RGBA, GL_UNSIGNED_BYTE, data );
		else
			glTexImage2D( GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB, w, h,
					0, GL_RGB, GL_UNSIGNED_BYTE, data );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
		glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

		free( data );
		return id;
	}
}

uint gi_img_load_bmp( const char *file, gi_raw_img *raw )
{
	fprintf( stderr, "%s:%i [ERROR] fixme: where is bmp loader? \n",
						__FILE__, __LINE__ );
	return 0;
}

void gi_img_delete( uint id, gi_raw_img *raw )
{
	if ( glIsTexture( id ) )
		glDeleteTextures( 1, &id );
	if ( raw != NULL )
		free( raw->raw );
}

void gi_ortho()
{
	f32 m[ 16 ] = {
		2.0f, 0.0f, 0.0f, 0.0f,
		0.0f, -2.0f, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, 0.0f,
		-1.0f, 1.0f, -1.0f, 1.0f };
	
	glMatrixMode( GL_PROJECTION );
	glLoadMatrixf( m );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

void gi_perspective(	f32 fov, f32 aspect,
			f32 znear, f32 zfar )
{
	f32 f = 1.0f / tanf( fov * ( M_PI / 360.0f ) );
	f32 m[ 16 ] = {
		f / aspect, 0.0f, 0.0f, 0.0f,
		0.0f, f, 0.0f, 0.0f,
		0.0f, 0.0f, ( zfar + znear ) / ( znear - zfar ), -1.0f,
		0.0f, 0.0f, ( zfar * znear * 2.0f ) / ( znear - zfar ), 0.0f };
	
	glMatrixMode( GL_PROJECTION );
	glLoadMatrixf( m );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}

void gi_lookat( const f32 *pos,
		const f32 *tgt,
		const f32 *up )
{
	f32 x[ 3 ], y[ 3 ], z[ 3 ];

	z[ 0 ] = - ( tgt[ 0 ] - pos[ 0 ] );
	z[ 1 ] = - ( tgt[ 1 ] - pos[ 1 ] );
	z[ 2 ] = - ( tgt[ 2 ] - pos[ 2 ] );

	f32 ins = ( f32 ) ( 1.0 / sqrt( z[ 0 ] * z[ 0 ] + z[ 1 ] * z[ 1 ] + z[ 2 ] * z[ 2 ] ) );
	z[ 0 ] *= ins;
	z[ 1 ] *= ins;
	z[ 2 ] *= ins;

	y[ 0 ] = up[ 0 ];
	y[ 1 ] = up[ 1 ];
	y[ 2 ] = up[ 2 ];

	x[ 0 ] = y[ 1 ] * z[ 2 ] - y[ 2 ] * z[ 1 ];
	x[ 1 ] = y[ 2 ] * z[ 0 ] - y[ 0 ] * z[ 2 ];
	x[ 2 ] = y[ 0 ] * z[ 1 ] - y[ 1 ] * z[ 0 ];

	ins = ( f32 ) ( 1.0 / sqrt( x[ 0 ] * x[ 0 ] + x[ 1 ] * x[ 1 ] + x[ 2 ] * x[ 2 ] ) );
	x[ 0 ] *= ins;
	x[ 1 ] *= ins;
	x[ 2 ] *= ins;

	y[ 0 ] = z[ 1 ] * x[ 2 ] - z[ 2 ] * x[ 1 ];
	y[ 1 ] = z[ 2 ] * x[ 0 ] - z[ 0 ] * x[ 2 ];
	y[ 2 ] = z[ 0 ] * x[ 1 ] - z[ 1 ] * x[ 0 ];

	ins = ( f32 ) ( 1.0 / sqrt( y[ 0 ] * y[ 0 ] + y[ 1 ] * y[ 1 ] + y[ 2 ] * y[ 2 ] ) );
	y[ 0 ] *= ins;
	y[ 1 ] *= ins;
	y[ 2 ] *= ins;

	f32 m[ 16 ] = {
		x[ 0 ], y[ 0 ], z[ 0 ], 0.0f,
		x[ 1 ], y[ 1 ], z[ 1 ], 0.0f,
		x[ 2 ], y[ 2 ], z[ 2 ], 0.0f,
		- ( pos[ 0 ] * x[ 0 ] + pos[ 1 ] * x[ 1 ] + pos[ 2 ] * x[ 2 ] ),
		- ( pos[ 0 ] * y[ 0 ] + pos[ 1 ] * y[ 1 ] + pos[ 2 ] * y[ 2 ] ),
		- ( pos[ 0 ] * z[ 0 ] + pos[ 1 ] * z[ 1 ] + pos[ 2 ] * z[ 2 ] ), 1.0f
	};

	glLoadMatrixf( m );
}
