#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "../def.h"
#include "map.h"
#include "config.h"

#define S_CONF_LINE_MAX		256

s_config *s_config_read( const char *file )
{
	FILE *fd = fopen( file, "r" );

	if ( fd == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] cannot read config \"%s\" \n",
				__FILE__, __LINE__, file );
		return NULL;
	}

	char buf[ S_CONF_LINE_MAX ];
	s_config *conf = malloc( sizeof( s_config ) );
	conf->recs = s_map_create( ( s_map_compare ) strcmp );
	conf->fname = strdup( file );

	while ( fgets( buf, S_CONF_LINE_MAX, fd ) != NULL )
	{
		int i, len = strlen( buf ), p = 0 ;
		for ( i = 0; i < len - p; )
		{
			if ( ( buf[ i ] == ' ' ) ||
					( buf[ i ] == '\t' ) ||
					( buf[ i ] == '\n' ) )
				p++;
			else
				i++;
			buf[ i ] = buf[ i + p ];
		}

		buf[ len - p ] = '\x00';

		if ( buf[ 0 ] != '#' )
		{
			char *name = buf;
			char *val = strchr( buf, '=' );

			if ( val == NULL )
			{
				fprintf( stderr, "%s:%i [ERROR] ivalid syntax: \"%s\" \n",
						__FILE__, __LINE__, buf );
				fclose( fd );
				return conf;
			}
			else
				*val++ = '\x00';

			char *name_cpy = strdup( name );
			char *ret = s_map_set( conf->recs, name_cpy, strdup( val ) );
			if ( ret != NULL )
			{
				free( name_cpy );
				free( ret );
			}
		}
	}

	fclose( fd );

	return conf;
}

void s_config_write( s_config *conf )
{
	FILE *fd = fopen( conf->fname, "w" );

	if ( fd == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] cannot write config \"%s\" \n",
				__FILE__, __LINE__, conf->fname );
		return ;
	}

	fprintf( fd, "# config file\n" );
	s_map_node *node = s_map_get_left( conf->recs );

	while ( node != NULL )
	{
		fprintf( fd, "%s=%s\n", ( char* ) node->index, ( char* ) node->data );
		node = s_map_move_right( conf->recs, node );
	}

	fclose( fd );
}

s_config *s_config_create( const char *file )
{
	s_config *conf = malloc( sizeof( s_config ) );
	conf->recs = s_map_create( ( s_map_compare ) strcmp );
	conf->fname = strdup( file );
	return conf;
}

void s_config_delete( s_config *conf )
{
	s_map_node *node = s_map_get_right( conf->recs );
	while ( node != NULL )
	{
		free( ( void* ) node->index );
		free( node->data );
		node = s_map_move_left( conf->recs, node );
	}

	s_map_delete( conf->recs );
	free( conf->fname );
	free( conf );
}

static void log_miss( const char *name )
{
	fprintf( stderr, "%s:%i [ERROR] record not found for \"%s\" \n",
				__FILE__, __LINE__, name );
}

const char *s_config_get_string( s_config *conf, const char *name )
{
	s_map_node *node = s_map_get( conf->recs, name );
	if ( node == NULL )
	{
		log_miss( name );
		return "";
	}
	else
		return node->data;
}

int s_config_get_int( s_config *conf, const char *name )
{
	s_map_node *node = s_map_get( conf->recs, name );
	if ( node == NULL )
	{
		log_miss( name );
		return 0;
	}
	else
		return strtol( node->data, NULL, 10 );
}

f64 s_config_get_float( s_config *conf, const char *name )
{
	s_map_node *node = s_map_get( conf->recs, name );
	if ( node == NULL )
	{
		log_miss( name );
		return 0.0;
	}
	else
		return strtod( node->data, NULL );
}

void s_config_set_string( s_config *conf, const char *name, const char *val )
{
	char *name_cpy = strdup( name );
	char *ret = s_map_set( conf->recs, name_cpy, strdup( val ) );
	if ( ret != NULL )
	{
		free( name_cpy );
		free( ret );
	}
}

void s_config_set_int( s_config *conf, const char *name, int val )
{
	char buf[ 40 ];
	sprintf( buf, "%i", val );
	char *name_cpy = strdup( name );
	char *ret = s_map_set( conf->recs, name_cpy, strdup( buf ) );
	if ( ret != NULL )
	{
		free( name_cpy );
		free( ret );
	}
}

void s_config_set_float( s_config *conf, const char *name, f64 val )
{
	char buf[ 40 ];
	sprintf( buf, "%lf", val );
	char *name_cpy = strdup( name );
	char *ret = s_map_set( conf->recs, name_cpy, strdup( buf ) );
	if ( ret != NULL )
	{
		free( name_cpy );
		free( ret );
	}
}

