#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "../src/def.h"

#define NAME		"meshconv: "
#define MAX_FILES	64

#define EQDIST		0.01f		/* distance between vertexes */
#define EQDIST_UV	0.001f		/* uv distance */
#define EQDIST_NO	4		/* normal distance */
#define VERTEQ(v0,v1)	( ( fabsf( v0.x - v1.x ) < EQDIST ) && \
			( fabsf( v0.y - v1.y ) < EQDIST ) && \
			( fabsf( v0.z - v1.z ) < EQDIST ) && \
			( fabsf( v0.u - v1.u ) < EQDIST_UV ) && \
			( fabsf( v0.v - v1.v ) < EQDIST_UV ) && \
			( abs( v0.nx - v1.nx ) < EQDIST_NO ) && \
			( abs( v0.ny - v1.ny ) < EQDIST_NO ) && \
			( abs( v0.nz - v1.nz ) < EQDIST_NO ) )

typedef struct {
	f32 x;
	f32 y;
	f32 z;
	f32 u;
	f32 v;
	i8 nx;
	i8 ny;
	i8 nz;
	i8 _z;
} vertex;

typedef struct {
	ui16 index_n;
	ui16 vertex_n;
	ui16 *idata;
	vertex *vdata;
} mesh;

static mesh* mesh_read( const char *file )
{
	mesh *m = malloc( sizeof( mesh ) );
	
	FILE *f = fopen( file, "rb" );
	
	if ( f == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] can't open file \"%s\" \n",
					__FILE__, __LINE__, file );
		exit( EXIT_FAILURE );
	}
	
	fread( &m->index_n, 2, 1, f );
	fread( &m->vertex_n, 2, 1, f );
	
	m->idata = malloc( m->index_n * 2 );
	fread( m->idata, 2 * m->index_n, 1, f );
	
	m->vdata = malloc( m->vertex_n * sizeof( vertex ) );
	fread( m->vdata, sizeof( vertex ) * m->vertex_n, 1, f );
	
	fclose( f );
	
	return m;
}

static void mesh_write( mesh *m, const char *file )
{
	FILE *f = fopen( file, "wr" );
	
	if ( f == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] can't open file \"%s\" \n",
					__FILE__, __LINE__, file );
		exit( EXIT_FAILURE );
	}
	
	fwrite( &m->index_n, 2, 1, f );
	fwrite( &m->vertex_n, 2, 1, f );
	fwrite( m->idata, 2 * m->index_n, 1, f );
	fwrite( m->vdata, sizeof( vertex ) * m->vertex_n, 1, f );
	
	fclose( f );
	
	free( m->idata );
	free( m->vdata );
	free( m );
}

static void print_help()
{
	printf( "\n" );
	printf( "usage: meshconv -[in..] file [file(0)] ... [file(n)]\n\n" );
	printf( " -help		: display this text\n" );
	printf( " -i		: recalculate mesh indexes\n" );
	printf( " -o		: attempt to optimize indexes\n" );
	printf( " -n		: recalculate mesh normals\n" );
	printf( " -f		: print mesh info\n" );
	printf( " -verbose	: to be verbose\n" );
	printf( "\n" );
}

int main( int argc, char *argv[] )
{
	int i = 1;
	int file_n = 0;
	char **file_list = malloc( sizeof( int* ) * MAX_FILES );
	
	bool reindex = FALSE;		/* -i */
	bool optimize = FALSE;		/* -o */
	bool minfo = FALSE;		/* -f */
	bool verbose = FALSE;		/* -verbose */
	
	if ( argc > 1 )
	{
		do {
			if ( argv[ i ][ 0 ] == '-' )
			{
				if ( !strcmp( argv[ i ], "-help" ) ||
					!strcmp( argv[ i ], "--help" ) )
				{
					print_help();
					exit( EXIT_SUCCESS );
				}
				else
				if ( !strcmp( argv[ i ], "-verbose" ) )
				{
					verbose = TRUE;
				}
				else
				{
					int j = 1;
					while ( argv[ i ][ j ] != 0 )
					{
						if ( argv[ i ][ j ] == 'i' )
							reindex = TRUE;
						else
						if ( argv[ i ][ j ] == 'o' )
							optimize = TRUE;
						else
						if ( argv[ i ][ j ] == 'f' )
							minfo = TRUE;
						j++;
					}
				}
			}
			else
			{
				if ( file_n < MAX_FILES )
					file_list[ file_n++ ] = strdup( argv[ i ] );
			}
			
			i++;
		} while ( i < argc );
	}
	else
	{
		printf( NAME "too few parameters, use --help \n" );
		exit( EXIT_FAILURE );
	}
	
	if ( ( reindex == FALSE ) &&
		( optimize == FALSE ) &&
		( minfo == FALSE ) )
	{
		printf( NAME "no defined actions \n" );
		exit( EXIT_FAILURE );
	}
	
	if ( file_n == 0 )
	{
		printf( NAME "no files \n" );
		exit( EXIT_FAILURE );
	}
	
	for ( i = 0; i < file_n; i++ )
	{
		if ( verbose )
		{
			printf( NAME "reading file \"%s\" \n", file_list[ i ] );
			fflush( stdout );
		}
		
		mesh *m = mesh_read( file_list[ i ] );
		
		if ( reindex )
		{
			if ( verbose )
			{
				printf( NAME "*** indexing ... \n" );
				printf( NAME "start info \n" );
				printf( NAME " index_n = %i \n", m->index_n );
				printf( NAME " vertex_n = %i \n",  m->vertex_n );
				fflush( stdout );
			}
			
			int j, k, lvs_n = m->index_n;
			vertex *lvs = malloc( lvs_n * sizeof( vertex ) );
			
			for ( j = 0; j < lvs_n; j++ )
			{
				memcpy( &lvs[ j ], &m->vdata[ m->idata[ j ] ],
					sizeof( vertex ) );
			}
			
			m->index_n = 0;
			m->vertex_n = 0;
			m->idata = realloc( m->idata, lvs_n * 2 );
			m->vdata = realloc( m->vdata, lvs_n * sizeof( vertex ) );
			
			for ( j = 0; j < lvs_n; j++ )
			{
				bool add = TRUE;
				for ( k = m->vertex_n - 1; k >= 0; k-- )
					if ( VERTEQ( lvs[ j ], m->vdata[ k ] ) )
					{
						add = FALSE;
						m->idata[ m->index_n++ ] = k;
						break;
					}
				
				if ( add )
				{
					m->idata[ m->index_n++ ] = m->vertex_n;
					memcpy( &m->vdata[ m->vertex_n++ ],
						&lvs[ j ], sizeof( vertex ) );
				}
			}
			/*
			m->idata = realloc( m->idata, m->index_n * 2 );
			m->vdata = realloc( m->vdata, m->vertex_n * sizeof( vertex ) );
			*/
			free( lvs );
			
			if ( verbose )
			{
				printf( NAME "final info \n" );
				printf( NAME " vertex_n = %i \n",  m->vertex_n );
				fflush( stdout );
			}
		}
		
		if ( optimize )
		{
			//FIXME:
		}
		
		if ( minfo )
		{
			printf( NAME "*** mesh info \n" );
			printf( NAME " index_n = %i \n", m->index_n );
			printf( NAME " vertex_n = %i \n",  m->vertex_n );
			fflush( stdout );
		}
		
		if ( verbose )
		{
			printf( NAME "writing file \"%s\" \n", file_list[ i ] );
			fflush( stdout );
		}
		
		mesh_write( m, file_list[ i ] );
	}
	
	for ( i = 0; i < file_n; i++ )
		free( file_list[ i ] );
	free( file_list );
	
	return EXIT_SUCCESS;
}
